import { ChangeDetectorRef, Directive, Input, OnDestroy, OnInit, TemplateRef, ViewContainerRef, } from '@angular/core';
import { Subscription, } from 'rxjs';
import { createRenderAware } from '../core';
import { createTemplateManager, } from '../core/utils/template-manager_creator';
import { DEFAULT_STRATEGY_NAME, getStrategies, } from '../render-strategies/strategies/strategies-map';
/**
 * @Directive LetDirective
 *
 * @description
 *
 * The `*rxLet` directive serves a convenient way of binding observables to a view context. Furthermore, it helps
 * you structure view-related models into view context scope (DOM element's scope).
 *
 * Under the hood, it leverages a `RenderStrategy` which in turn takes care of optimizing the change detection
 * of your component. The `LetDirective` will render its template and manage change detection after it got an initial value.
 * So if the incoming `Observable` emits its value lazily (e.g. data coming from `Http`), your template will be
 * rendered lazily as well. This can very positively impact the initial render performance of your application.
 *
 *
 * ### Problems with `async` and `*ngIf`
 *
 * In Angular, a way of binding an observable to the view could look like that:
 * ```html
 * <ng-container *ngIf="observableNumber$ | async as n">
 *   <app-number [number]="n"></app-number>
 *   <app-number-special [number]="n"></app-number-special>
 * </ng-container>
 * ```
 *
 * The problem is that `*ngIf` interferes with rendering and in case of a `0` (a falsy value) the component
 * would be hidden. This issue doesn't concern the `LetDirective`.
 *
 * The `AsyncPipe` relies on the Zone to be present - it doesn't really trigger change detection by itself.
 * It marks the component and its children as dirty waiting for the Zone to trigger change detection. So, in case
 * you want to create a zone-less application, the `AsyncPipe` won't work as desired. `LetDirective` comes
 * with its own strategies to manage change detection every time a new notification is sent from
 * the bound Observable.
 *
 *
 * ### Features of `*rxLet`
 *
 * Included features for `*rxLet`:
 * - binding is always present. (see "Problems with `async` and `*ngIf`" section below)
 * - it takes away the multiple usages of the `async` or `push` pipe
 * - a unified/structured way of handling null and undefined
 * - triggers change-detection differently if `zone.js` is present or not (`ChangeDetectorRef.detectChanges` or
 *   `ChangeDetectorRef.markForCheck`)
 * - triggers change-detection differently if ViewEngine or Ivy is present (`ChangeDetectorRef.detectChanges` or
 *   `ɵdetectChanges`)
 * - distinct same values in a row (`distinctUntilChanged` operator),
 * - display custom templates for different observable notifications (suspense, next, error, complete)
 *
 *
 * ### Binding an Observable and using the view context
 *
 * The `*rxLet` directive takes over several things and makes it more convenient and save to work with streams in the
 * template:
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 *
 * <ng-container *rxLet="observableNumber$ as n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * ```
 *
 * In addition to that it provides us information from the whole observable context.
 * We can track the observables:
 * - next value
 * - error occurrence
 * - complete occurrence
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n; let e = $error, let c = $complete">
 *   <app-number [number]="n" *ngIf="!e && !c"></app-number>
 *   <ng-container *ngIf="e">
 *     There is an error: {{ e }}
 *   </ng-container>
 *   <ng-container *ngIf="c">
 *     Observable completed: {{ c }}
 *   </ng-container>
 * </ng-container>
 * ```
 *
 *
 * ### Using the template-binding
 *
 * You can also use template anchors and display template's content for different observable states:
 * - on complete
 * - on error
 * - on suspense - before the first value is emitted
 *
 * ```html
 * <ng-container
 *   *rxLet="
 *     observableNumber$;
 *     let n;
 *     error: error;
 *     complete: complete;
 *     suspense: suspense;
 *   "
 * >
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * <ng-template #error>ERROR</ng-template>
 * <ng-template #complete>COMPLETE</ng-template>
 * <ng-template #suspense>SUSPENSE</ng-template>
 * ```
 *
 * Internally, `*rxLet` is using a simple "view memoization" - it caches all anchored template references and re-uses
 * them whenever the observable notification (next/error/complete) is sent. Then, it only updates the context
 * (e.g. a value from the observable) in the view.
 *
 * @docsCategory LetDirective
 * @docsPage LetDirective
 * @publicApi
 */
var LetDirective = /** @class */ (function () {
    /**
     * @internal
     */
    function LetDirective(cdRef, nextTemplateRef, viewContainerRef) {
        var _this = this;
        this.nextTemplateRef = nextTemplateRef;
        this.viewContainerRef = viewContainerRef;
        this.subscription = new Subscription();
        this.resetObserver = {
            next: function () {
                _this.displayInitialView();
                _this.templateManager.updateViewContext({
                    $implicit: undefined,
                    rxLet: undefined,
                    $error: false,
                    $complete: false,
                });
            },
        };
        this.updateObserver = {
            next: function (value) {
                _this.templateManager.displayView('rxNext');
                _this.templateManager.updateViewContext({
                    $implicit: value,
                    rxLet: value,
                });
            },
            error: function (error) {
                // fallback to rxNext when there's no template for rxError
                _this.templateManager.hasTemplateRef('rxError')
                    ? _this.templateManager.displayView('rxError')
                    : _this.templateManager.displayView('rxNext');
                _this.templateManager.updateViewContext({
                    $error: true,
                });
            },
            complete: function () {
                // fallback to rxNext when there's no template for rxComplete
                _this.templateManager.hasTemplateRef('rxComplete')
                    ? _this.templateManager.displayView('rxComplete')
                    : _this.templateManager.displayView('rxNext');
                _this.templateManager.updateViewContext({
                    $complete: true,
                });
            },
        };
        this.displayInitialView = function () {
            // display "suspense" template if provided
            if (_this.templateManager.hasTemplateRef('rxSuspense')) {
                _this.templateManager.displayView('rxSuspense');
            }
        };
        this.strategies = getStrategies({ cdRef: cdRef });
        this.templateManager = createTemplateManager(this.viewContainerRef, {
            $implicit: undefined,
            rxLet: undefined,
            $error: false,
            $complete: false,
        });
        this.renderAware = createRenderAware({
            strategies: this.strategies,
            resetObserver: this.resetObserver,
            updateObserver: this.updateObserver,
        });
        this.renderAware.nextStrategy(DEFAULT_STRATEGY_NAME);
    }
    Object.defineProperty(LetDirective.prototype, "rxLet", {
        /**
         * @description
         * The Observable to be bound to the context of a template.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         *
         * @param potentialObservable
         */
        set: function (potentialObservable) {
            this.renderAware.nextPotentialObservable(potentialObservable);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "strategy", {
        /**
         * @description
         * The rendering strategy to be used when rendering with the reactive context within a template.
         * Use it to dynamically manage your rendering strategy. You can switch the strategies
         * imperatively (with a string) or by bounding an Observable.
         * The default strategy is `'local'`.
         *
         * @example
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *     <ng-container *rxLet="hero$; let hero; strategy: 'local'">
         *       <app-hero [hero]="hero"></app-hero>
         *     </ng-container>
         *   `
         * })
         * export class AppComponent {
         *   strategy = 'local';
         * }
         *
         * // OR
         *
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *     <ng-container *rxLet="hero$; let hero; strategy: strategy$">
         *       <app-hero [hero]="hero"></app-hero>
         *     </ng-container>
         *   `
         * })
         * export class AppComponent {
         *   strategy$ = new BehaviorSubject('local');
         * }
         *
         * @param strategy
         * @see {@link strategies}
         */
        set: function (strategy) {
            this.renderAware.nextStrategy(strategy || DEFAULT_STRATEGY_NAME);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "rxLetComplete", {
        /**
         * @description
         * A template to show if the bound Observable is in "complete" state.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero; complete: completeTemplate">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         * <ng-template #completeTemplate>
         *   <mat-icon>thumb_up</mat-icon>
         * </ng-template>
         *
         * @param templateRef
         */
        set: function (templateRef) {
            this.templateManager.addTemplateRef('rxComplete', templateRef);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "rxLetError", {
        /**
         * @description
         * A template to show if the bound Observable is in "error" state.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero; error: errorTemplate">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         * <ng-template #errorTemplate>
         *   <mat-icon>thumb_down</mat-icon>
         * </ng-template>
         *
         * @param templateRef
         */
        set: function (templateRef) {
            this.templateManager.addTemplateRef('rxError', templateRef);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "rxLetSuspense", {
        /**
         * @description
         * A template to show before the first value is emitted from the bound Observable.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero; suspense: suspenseTemplate">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         * <ng-template #suspenseTemplate>
         *   <mat-progress-spinner></mat-progress-spinner>
         * </ng-template>
         *
         * @param templateRef
         */
        set: function (templateRef) {
            this.templateManager.addTemplateRef('rxSuspense', templateRef);
        },
        enumerable: false,
        configurable: true
    });
    /**
     * @internal
     */
    LetDirective.ngTemplateContextGuard = function (dir, ctx) {
        return true;
    };
    /**
     * @internal
     */
    LetDirective.prototype.ngOnInit = function () {
        this.templateManager.addTemplateRef('rxNext', this.nextTemplateRef);
        this.displayInitialView();
        this.subscription = this.renderAware.subscribe();
    };
    /**
     * @internal
     */
    LetDirective.prototype.ngOnDestroy = function () {
        this.subscription.unsubscribe();
        this.templateManager.destroy();
    };
    LetDirective.ctorParameters = function () { return [
        { type: ChangeDetectorRef },
        { type: TemplateRef },
        { type: ViewContainerRef }
    ]; };
    LetDirective.decorators = [
        { type: Directive, args: [{
                    selector: '[rxLet]',
                    exportAs: 'renderNotifier',
                },] }
    ];
    LetDirective.ctorParameters = function () { return [
        { type: ChangeDetectorRef },
        { type: TemplateRef },
        { type: ViewContainerRef }
    ]; };
    LetDirective.propDecorators = {
        rxLet: [{ type: Input }],
        strategy: [{ type: Input, args: ['rxLetStrategy',] }],
        rxLetComplete: [{ type: Input }],
        rxLetError: [{ type: Input }],
        rxLetSuspense: [{ type: Input }]
    };
    return LetDirective;
}());
export { LetDirective };
//# sourceMappingURL=let.directive.js.map