import { Observable, Subscription, ReplaySubject, of, from, BehaviorSubject, Subject } from 'rxjs';
import { __read, __assign } from 'tslib';
import { distinctUntilChanged, switchMap, map, tap, shareReplay, filter, withLatestFrom, mergeAll } from 'rxjs/operators';
import { ElementRef, Directive, Input, NgModule, ɵmarkDirty, ɵdetectChanges, ChangeDetectorRef, TemplateRef, ViewContainerRef, Optional, Pipe } from '@angular/core';

/*
 * createPropertiesWeakMap
 *
 * @param getDefaults: (o: O) => P
 * Example:
 *
 * export interface Properties {
 *   isCoalescing: boolean;
 * }
 *
 * const obj: object = {
 *   foo: 'bar',
 *   isCoalescing: 'weakMap version'
 * };
 *
 * const getDefaults = (ctx: object): Properties => ({isCoalescing: false});
 * const propsMap = createPropertiesWeakMap<object, Properties>(getDefaults);
 *
 * console.log('obj before:', obj);
 * // {foo: "bar", isCoalescing: "weakMap version"}
 * console.log('props before:', propsMap.getProps(obj));
 * // {isCoalescing: "weakMap version"}
 *
 * propsMap.setProps(obj, {isCoalescing: true});
 * console.log('obj after:', obj);
 * // {foo: "bar", isCoalescing: "weakMap version"}
 * console.log('props after:', propsMap.getProps(obj));
 * // {isCoalescing: "true"}
 * */
function createPropertiesWeakMap(getDefaults) {
    var propertyMap = new WeakMap();
    return {
        getProps: getProperties,
        setProps: setProperties,
    };
    function getProperties(ctx) {
        var defaults = getDefaults(ctx);
        var propertiesPresent = propertyMap.get(ctx);
        var properties;
        if (propertiesPresent !== undefined) {
            properties = propertiesPresent;
        }
        else {
            properties = {};
            Object.entries(defaults).forEach(function (_a) {
                var _b = __read(_a, 2), prop = _b[0], value = _b[1];
                properties[prop] = hasKey(ctx, prop) ? ctx[prop] : value;
            });
            propertyMap.set(ctx, properties);
        }
        return properties;
    }
    function setProperties(ctx, props) {
        var properties = getProperties(ctx);
        Object.entries(props).forEach(function (_a) {
            var _b = __read(_a, 2), prop = _b[0], value = _b[1];
            properties[prop] = value;
        });
        propertyMap.set(ctx, properties);
        return properties;
    }
    function hasKey(ctx, property) {
        return ctx[property] != null;
    }
}

var coalescingManager = createCoalesceManager();
var ɵ0 = function (ctx) { return ({
    numCoalescingSubscribers: 0,
}); };
var coalescingContextPropertiesMap = createPropertiesWeakMap(ɵ0);
/**
 * @describe createCoalesceManager
 *
 * returns a
 * Maintains a weak map of component references ans flags
 * them if the coalescing process is already started for them.
 *
 * Used in render aware internally.
 */
function createCoalesceManager() {
    return {
        remove: removeWork,
        add: addWork,
        isCoalescing: isCoalescing,
    };
    // Increments the number of subscriptions in a scope e.g. a class instance
    function removeWork(scope) {
        if (scope === void 0) { scope = {}; }
        var numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers -
            1;
        coalescingContextPropertiesMap.setProps(scope, {
            numCoalescingSubscribers: numCoalescingSubscribers,
        });
    }
    // Decrements the number of subscriptions in a scope e.g. a class instance
    function addWork(scope) {
        if (scope === void 0) { scope = {}; }
        var numCoalescingSubscribers = coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers +
            1;
        coalescingContextPropertiesMap.setProps(scope, {
            numCoalescingSubscribers: numCoalescingSubscribers,
        });
    }
    // Checks if anybody else is already coalescing atm
    function isCoalescing(scope) {
        if (scope === void 0) { scope = {}; }
        return (coalescingContextPropertiesMap.getProps(scope).numCoalescingSubscribers >
            0);
    }
}

/**
 * @description
 * Limits the number of synchronous emitted a value from the source Observable to
 * one emitted value per
 *   [`AnimationFrame`](https://developer.mozilla.org/en-US/docs/Web/API/Window/requestAnimationFrame), then repeats
 *   this process for every tick of the browsers event loop.
 *
 * The coalesce operator is based on the [throttle](https://rxjs-dev.firebaseapp.com/api/operators/throttle) operator.
 * In addition to that is provides emitted values for the trailing end only, as well as maintaining a context to scope
 *   coalescing.
 *
 * @param {function(value: T): SubscribableOrPromise} durationSelector - A function
 * that receives a value from the source Observable, for computing the silencing
 * duration for each source value, returned as an Observable or a Promise.
 * It defaults to `requestAnimationFrame` as durationSelector.
 * @param {Object} config - A configuration object to define `leading` and `trailing` behavior and the context object.
 * Defaults to `{ leading: false, trailing: true }`. The default scoping is per subscriber.
 * @return {Observable<T>} An Observable that performs the coalesce operation to
 * limit the rate of emissions from the source.
 *
 * @usageNotes
 * Emit clicks at a rate of at most one click per second
 * ```typescript
 * import { fromEvent, animationFrames } from 'rxjs';
 * import { coalesce } from 'ngRx/component';
 *
 * const clicks = fromEvent(document, 'click');
 * const result = clicks.pipe(coalesce(ev => animationFrames));
 * result.subscribe(x => console.log(x));
 * ```
 */
function coalesceWith(durationSelector, scope) {
    var _scope = scope || {};
    return function (source) {
        var o$ = new Observable(function (observer) {
            var rootSubscription = new Subscription();
            rootSubscription.add(source.subscribe(createInnerObserver(observer, rootSubscription)));
            return rootSubscription;
        });
        return o$;
        function createInnerObserver(outerObserver, rootSubscription) {
            var actionSubscription;
            var latestValue;
            var tryEmitLatestValue = function () {
                coalescingManager.remove(_scope);
                if (!coalescingManager.isCoalescing(_scope)) {
                    outerObserver.next(latestValue);
                }
            };
            return {
                complete: function () {
                    if (actionSubscription) {
                        tryEmitLatestValue();
                    }
                    outerObserver.complete();
                },
                error: function (error) { return outerObserver.error(error); },
                next: function (value) {
                    latestValue = value;
                    if (!actionSubscription) {
                        coalescingManager.add(_scope);
                        actionSubscription = durationSelector.subscribe({
                            next: function () {
                                tryEmitLatestValue();
                                actionSubscription = undefined;
                            },
                            complete: function () {
                                if (actionSubscription) {
                                    tryEmitLatestValue();
                                    actionSubscription = undefined;
                                }
                            },
                        });
                        rootSubscription.add(actionSubscription);
                    }
                },
            };
        }
    };
}

/**
 * @description
 *
 * A fallback for the new `globalThis` reference.
 *
 *  It should be used to replace `window` due to different environments in:
 *  - SSR (Server Side Rendering)
 *  - Tests
 *  - Browser
 *
 *  @return {globalThis} - A reference to globalThis. `window` in the Browser.
 */
function getGlobalThis() {
    return (window || self || globalThis);
}

/**
 * envZonePatched
 *
 * @description
 *
 * This function checks the window object `zone.js` was instantiated.
 * If so, the `window` object maintains a property named `Zone`.
 *
 * Here how Angular checks it: https://github.com/angular/angular/blob/master/packages/core/src/zone/ng_zone.ts#L123
 *
 * @return {boolean} - true if `zone.js` patched global APIs.
 *
 */
function envZonePatched() {
    return getGlobalThis().Zone !== undefined;
}
/**
 * apiZonePatched
 *
 * @description
 *
 * This function checks if a specific Browser API is patched by `zone.js`.
 *
 * @param name {string} - The name of the API to check.
 * @return {boolean} - true if `zone.js` patched the API in question.
 *
 */
function apiZonePatched(name) {
    // if symbol is present, zone patched the API
    return getGlobalThis()['__zone_symbol__' + name] !== undefined;
}
var zoneDetectionCache = new WeakMap();
/**
 * isNgZone
 *
 * @description
 *
 * This function takes an instance of a class which implements the NgZone interface and checks if
 * its `runOutsideAngular()` function calls `apply()` on the function passed as parameter. This
 * means the Angular application that instantiated this service assumes it runs in a ZoneLess
 * environment, and therefore it's change detection will not be triggered by zone related logic.
 *
 * However, keep in mind this does not mean `zone.js` is not present.
 * The environment could still run in ZoneFull mode even if Angular turned it off.
 * Consider the situation of a Angular element configured for ZoneLess
 * environments is used in an Angular application relining on the zone mechanism.
 *
 * @param instance {Class Instance} - The instance to check for constructor name of `NgZone`.
 * @return {boolean} - true if instance is of type `NgZone`.
 *
 */
function isNgZone(instance) {
    var cachedValue = zoneDetectionCache.get(instance);
    if (cachedValue !== undefined) {
        return cachedValue;
    }
    var calledApply = false;
    function fn() { }
    fn.apply = function () { return (calledApply = true); };
    instance.runOutsideAngular(fn);
    zoneDetectionCache.set(instance, calledApply);
    return calledApply;
}
/**
 * isNoopNgZone
 *
 *@description
 *
 * This function takes any instance of a class and checks
 * if the constructor name is equal to `NoopNgZone`.
 *
 * For more detailed information read the description of [isNgZone](#isngzone).
 *
 * @param instance {Class Instance} - The instance to check for constructor name of `NoopNgZone`.
 * @return {boolean} - true if instance is of type `NoopNgZone`.
 *
 */
function isNoopNgZone(instance) {
    return !isNgZone(instance);
}

/** A shared promise instance to cause a delay of one microtask */
var resolvedPromise = null;
function getUnpatchedResolvedPromise() {
    resolvedPromise =
        resolvedPromise ||
            (apiZonePatched('Promise')
                ? getGlobalThis().__zone_symbol__Promise.resolve()
                : Promise.resolve());
    return resolvedPromise;
}

/**
 * getZoneUnPatchedApi
 *
 * @description
 *
 * This function returns the zone un-patched API for the a specific Browser API.
 * If no element is passed the window is used instead
 *
 * @param name {string} - The name of the API to check.
 * @param elem {any} - The elem to get un-patched API from.
 * @return {Function} - The zone un-patched API in question.
 *
 */
function getZoneUnPatchedApi(name, elem) {
    elem = elem || getGlobalThis();
    return apiZonePatched(name) ? elem['__zone_symbol__' + name] : elem[name];
}

function staticCoalesce(work, durationSelector, scope, abC) {
    if (scope === void 0) { scope = {}; }
    if (abC === void 0) { abC = new AbortController(); }
    var sub;
    if (!coalescingManager.isCoalescing(scope)) {
        coalescingManager.add(scope);
        sub = durationSelector.subscribe(function () {
            tryExecuteWork();
        });
        var abortHandler_1 = function () {
            sub.unsubscribe();
            abC.signal.removeEventListener('abort', abortHandler_1, false);
        };
        abC.signal.addEventListener('abort', abortHandler_1, false);
    }
    return abC;
    // =====
    function tryExecuteWork() {
        coalescingManager.remove(scope);
        if (!coalescingManager.isCoalescing(scope)) {
            return work();
        }
    }
}

var PostTaskSchedulerPriority;
(function (PostTaskSchedulerPriority) {
    PostTaskSchedulerPriority["background"] = "background";
    PostTaskSchedulerPriority["userBlocking"] = "user-blocking";
    PostTaskSchedulerPriority["userVisible"] = "user-visible";
})(PostTaskSchedulerPriority || (PostTaskSchedulerPriority = {}));
var postTaskScheduler = typeof window !== 'undefined'
    ? window.scheduler || {
        postTask: function (options) {
            var start = Date.now();
            return new Promise(function (resolve) {
                setTimeout(function () {
                    console.error('postTask not implemented. Use setTimeout as fallback');
                    resolve();
                }, 1);
            });
        },
    }
    : function () { };

var timeoutTick = function () {
    return new Observable(function (subscriber) {
        var id = window.__zone_symbol__setTimeout(function () {
            subscriber.next(0);
            subscriber.complete();
        });
        return function () {
            window.__zone_symbol__clearTimeout(id);
        };
    });
};

var cancelIdleCallback = typeof window !== 'undefined'
    ? window.cancelIdleCallback ||
        function (idleId) {
            console.warn('Fake cancelIdleCallback used');
            clearTimeout(idleId);
        }
    : function () { };
var requestIdleCallback = typeof window !== 'undefined'
    ? window.requestIdleCallback ||
        function (cb) {
            console.warn('Fake requestIdleCallback used');
            var start = Date.now();
            return setTimeout(function () {
                cb({
                    didTimeout: false,
                    timeRemaining: function () {
                        return Math.max(0, 50 - (Date.now() - start));
                    },
                });
            }, 1);
        }
    : function () { };

var idleCallbackTick = function () {
    return new Observable(function (subscriber) {
        var id = requestIdleCallback(function () {
            subscriber.next(0);
            subscriber.complete();
        });
        return function () { return cancelIdleCallback(id); };
    });
};

var postTaskTick = function (options) {
    return new Observable(function (subscription) {
        postTaskScheduler
            .postTask(function () {
        }, options)
            .then(function () {
            subscription.next(0);
            subscription.complete();
        });
        return function () {
        };
    });
};

/**
 * RenderAware
 *
 * @description
 * This function returns an object that holds all the shared logic for the push pipe and the let directive
 * responsible for change detection
 * If you extend this class you need to implement how the update of the rendered value happens.
 * Also custom behaviour is something you need to implement in the extending class
 */
function createRenderAware(cfg) {
    var strategyName$ = new ReplaySubject(1);
    var currentStrategy;
    var strategy$ = strategyName$.pipe(distinctUntilChanged(), switchMap(function (stringOrObservable) {
        return typeof stringOrObservable === 'string'
            ? of(stringOrObservable)
            : stringOrObservable;
    }), map(function (strategy) {
        var s = cfg.strategies[strategy];
        if (!!s) {
            return s;
        }
        throw new Error("Strategy " + strategy + " does not exist.");
    }), tap(function (s) { return (currentStrategy = s); }), 
    // do not repeat the steps before for each subscriber
    shareReplay({ bufferSize: 1, refCount: true }));
    var observablesFromTemplate$ = new ReplaySubject(1);
    var valuesFromTemplate$ = observablesFromTemplate$.pipe(distinctUntilChanged());
    var firstTemplateObservableChange = true;
    var renderingEffect$ = valuesFromTemplate$.pipe(
    // handle null | undefined assignment and new Observable reset
    map(function (observable$) {
        if (observable$ === null) {
            return of(null);
        }
        if (!firstTemplateObservableChange) {
            cfg.resetObserver.next();
            if (observable$ === undefined) {
                return of(undefined);
            }
        }
        firstTemplateObservableChange = false;
        return observable$;
    }), 
    // forward only observable values
    filter(function (o$) { return o$ !== undefined; }), distinctUntilChanged(), switchMap(function (o$) {
        return o$
            // Added behavior will get applied to the observable in `renderWithLatestStrategy`
            .pipe(
        // Forward only distinct values
        distinctUntilChanged(), 
        // Update completion, error and next
        tap(cfg.updateObserver), renderWithLatestStrategy(strategy$));
    }));
    return {
        nextPotentialObservable: function (value) {
            observablesFromTemplate$.next(value);
        },
        nextStrategy: function (nextConfig) {
            strategyName$.next(nextConfig);
        },
        activeStrategy$: strategy$,
        subscribe: function () {
            return new Subscription().add(renderingEffect$.subscribe());
        },
    };
}
function renderWithLatestStrategy(strategyChanges$) {
    return function (o$) {
        return o$.pipe(handleErrorAndComplete(), withLatestFrom(strategyChanges$), 
        // always use latest strategy on value change
        switchMap(function (_a) {
            var _b = __read(_a, 2), renderValue = _b[0], strategy = _b[1];
            return of(renderValue).pipe(strategy.rxScheduleCD);
        }));
    };
    function handleErrorAndComplete() {
        return function (o$) {
            return new Observable(function (subscriber) {
                var subscription = o$.subscribe({
                    next: function (val) { return subscriber.next(val); },
                    // make "error" and "complete" notifications comply with `rxScheduleCD`
                    error: function (err) {
                        console.error(err);
                        subscriber.next();
                    },
                    complete: function () { return subscriber.next(); },
                });
                return function () { return subscription.unsubscribe(); };
            });
        };
    }
}

var animationFrameTick = function () {
    return new Observable(function (subscriber) {
        var id = getZoneUnPatchedApi('requestAnimationFrame')(function () {
            subscriber.next(0);
            subscriber.complete();
        });
        return function () {
            getZoneUnPatchedApi('cancelAnimationFrame')(id);
        };
    });
};

// @NOTICE replace logic with 7v handling of promises in RxJS
var promiseTick = function () {
    return new Observable(function (subscriber) {
        var cancelled = false;
        getUnpatchedResolvedPromise()
            .then(function () {
            if (!cancelled) {
                subscriber.next(0);
                subscriber.complete();
            }
        })
            .catch(function (e) {
            subscriber.error(e);
        });
        return function () {
            cancelled = true;
            subscriber.complete();
        };
    });
};

var priorityTickMap = {
    animationFrame: animationFrameTick(),
    Promise: promiseTick(),
    // @deprecated This is still experimental
    setInterval: timeoutTick(),
    // @deprecated This is still experimental
    idleCallback: idleCallbackTick(),
    // @deprecated This is still experimental
    userBlocking: postTaskTick({
        priority: PostTaskSchedulerPriority.userBlocking,
    }),
    // @deprecated This is still experimental
    userVisible: postTaskTick({
        priority: PostTaskSchedulerPriority.userVisible,
    }),
    // @deprecated This is still experimental
    background: postTaskTick({ priority: PostTaskSchedulerPriority.background }),
};

function staticSchedule(work, priority, abC) {
    if (abC === void 0) { abC = new AbortController(); }
    // immediately execute work
    if (priority === false) {
        tryExecuteWork();
        return abC;
    }
    // schedule work
    var sub = priorityTickMap[priority].subscribe(function () { return tryExecuteWork(); }, function (error) { return console.error(error); }, 
    // on complete abort further executions
    function () { return abC.abort(); });
    var abortHandler = function () {
        sub.unsubscribe();
        abC.signal.removeEventListener('abort', abortHandler, false);
    };
    abC.signal.addEventListener('abort', abortHandler, false);
    return abC;
    // execute work and abort further executions
    function tryExecuteWork() {
        if (!abC.signal.aborted) {
            work();
            abC.abort();
        }
    }
}

function coalesceAndSchedule(work, priority, scope, abC) {
    if (scope === void 0) { scope = {}; }
    if (abC === void 0) { abC = new AbortController(); }
    var durationSelector = from(getUnpatchedResolvedPromise());
    var scheduledWork = function () { return staticSchedule(work, priority, abC); };
    var coalesceAbC = staticCoalesce(scheduledWork, durationSelector, scope, abC);
    var abortHandler = function () {
        coalesceAbC.abort();
        abC.signal.removeEventListener('abort', abortHandler, false);
    };
    abC.signal.addEventListener('abort', abortHandler, false);
    return abC;
}

var zonePatchedEvents = [
    'scroll',
    'mousedown',
    'mouseenter',
    'mouseleave',
    'mousemove',
    'mouseout',
    'mouseover',
    'mouseup',
    'load',
    'pointerup',
    'change',
    'blur',
    'focus',
    'click',
    'contextmenu',
    'drag',
    'dragend',
    'dragenter',
    'dragleave',
    'dragover',
    'dragstart',
    'drop',
    'input',
];

/**
 *
 * @description
 *
 * This function takes an elem and event and re-applies the listeners from the passed event to the
 * passed element with the zone un-patched version of it.
 *
 * @param elem {HTMLElement} - The elem to re-apply the listeners to.
 * @param event {string} - The name of the event from which to re-apply the listeners.
 *
 * @returns void
 */
function unpatchEventListener(elem, event) {
    var eventListeners = elem.eventListeners(event);
    // Return if no event listeners are present
    if (!eventListeners) {
        return;
    }
    var addEventListener = getZoneUnPatchedApi('addEventListener', elem).bind(elem);
    eventListeners.forEach(function (listener) {
        // Remove and reapply listeners with patched API
        elem.removeEventListener(event, listener);
        // Reapply listeners with un-patched API
        addEventListener(event, listener);
    });
}
/**
 * @Directive UnpatchEventsDirective
 *
 * @description
 *
 * The `unpatch` directive helps in partially migrating to zone-less apps as well as getting rid
 * of unnecessary renderings through zones `addEventListener` patches.
 * It can be used on any element you apply event bindings.
 *
 * The current way of binding events to the DOM is to use output bindings:
 *  ```html
 * <button (click)="doStuff($event)">click me</button>
 * ```
 *
 * The problem is that every event registered over `()` syntax, e.g. `(click)`
 * marks the component and all its ancestors as dirty and re-renders the whole component tree.
 * This is because zone.js patches the native browser API and whenever one of the patched APIs is used it re-renders.
 *
 * So even if your button is not related to a change that needs a re-render the app will re-render completely.
 * This leads to bad performance. This is especially helpful if you work with frequently fired events like 'mousemove'
 *
 * `unpatch` directive solves that problem.
 *
 * Included Features:
 *  - by default un-patch all registered listeners of the host it is applied on
 *  - un-patch only a specified set of registered event listeners
 *  - works zone independent (it directly checks the widow for patched APIs and un-patches them without the use of `runOutsideZone` which brings more performance)
 *  - Not interfering with any logic executed by the registered callback
 *
 * @usageNotes
 *
 * The `unpatch` directive can be used like shown here:
 * ```html
 * <button [unoatch] (click)="triggerSomeMethod($event)">click me</button>
 * <button [unoatch]="['mousemove']" (mousemove)="doStuff2($event)" (click)="doStuff($event)">click me</button>
 * ```
 *
 * @publicApi
 */
// tslint:disable-next-line:directive-selector
var UnpatchEventsDirective = /** @class */ (function () {
    function UnpatchEventsDirective(el) {
        this.el = el;
        this.subscription = new Subscription();
        this.events$ = new BehaviorSubject(zonePatchedEvents);
    }
    Object.defineProperty(UnpatchEventsDirective.prototype, "events", {
        set: function (value) {
            if (value && value.length > 0) {
                this.events$.next(value);
            }
            else {
                this.events$.next(zonePatchedEvents);
            }
        },
        enumerable: false,
        configurable: true
    });
    UnpatchEventsDirective.prototype.reapplyEventListenersZoneUnPatched = function (events) {
        var _this = this;
        events.forEach(function (ev) {
            unpatchEventListener(_this.el.nativeElement, ev);
        });
    };
    UnpatchEventsDirective.prototype.ngOnDestroy = function () {
        this.subscription.unsubscribe();
    };
    UnpatchEventsDirective.prototype.ngAfterViewInit = function () {
        var _this = this;
        this.subscription = this.events$
            .pipe(tap(function (eventList) { return _this.reapplyEventListenersZoneUnPatched(eventList); }))
            .subscribe();
    };
    UnpatchEventsDirective.ctorParameters = function () { return [
        { type: ElementRef }
    ]; };
    UnpatchEventsDirective.decorators = [
        { type: Directive, args: [{ selector: '[unpatch]' },] }
    ];
    UnpatchEventsDirective.ctorParameters = function () { return [
        { type: ElementRef }
    ]; };
    UnpatchEventsDirective.propDecorators = {
        events: [{ type: Input, args: ['unpatch',] }]
    };
    return UnpatchEventsDirective;
}());

var DECLARATIONS = [UnpatchEventsDirective];
var UnpatchEventsModule = /** @class */ (function () {
    function UnpatchEventsModule() {
    }
    UnpatchEventsModule.decorators = [
        { type: NgModule, args: [{
                    declarations: DECLARATIONS,
                    exports: DECLARATIONS,
                },] }
    ];
    return UnpatchEventsModule;
}());

/**
 * TemplateManager
 *
 * @description
 * This function returns an object that holds the logic for managing templates of a `ViewContainerRef`.
 * It abstracts `EmbeddedView` creation, `insert` calls and `ViewContext` updates.
 * Internally it creates template references lazily by combining caching logic and the `ViewContainerRef#detach` method.
 * The `TemplateManager` lets you re-use templates and insert views on-demand, as well as update the view context
 * (e.g. when a new observable notification is sent).
 *
 * @param viewContainerRef reference to a top-level view container where passed templates will be attached
 * @param initialViewContext initial view context state
 */
function createTemplateManager(viewContainerRef, initialViewContext) {
    var templateCache = new Map();
    var viewCache = new Map();
    var viewContext = __assign({}, initialViewContext);
    var activeView;
    return {
        hasTemplateRef: function (name) {
            return templateCache.has(name);
        },
        updateViewContext: function (viewContextSlice) {
            Object.entries(viewContextSlice).forEach(function (_a) {
                var _b = __read(_a, 2), key = _b[0], value = _b[1];
                viewContext[key] = value;
            });
        },
        addTemplateRef: function (name, templateRef) {
            assertTemplate(name, templateRef);
            if (!templateCache.has(name)) {
                templateCache.set(name, templateRef);
            }
            else {
                // @Notice We have to think through how this would work. We also call viewCache.set(name, newView); in insertEmbeddedView
                throw new Error('Updating an already existing Template is not supported at the moment.');
            }
        },
        displayView: function (name) {
            if (activeView !== name) {
                if (templateCache.has(name)) {
                    // Detach currently inserted view from the container
                    viewContainerRef.detach();
                    if (viewCache.has(name)) {
                        viewContainerRef.insert(viewCache.get(name));
                    }
                    else {
                        // Creates and inserts view to the view container
                        var newView = viewContainerRef.createEmbeddedView(templateCache.get(name), viewContext);
                        viewCache.set(name, newView);
                    }
                }
                else {
                    // @NOTICE this is here to cause errors and see in which situations we would throw.
                    // In CDK it should work different.
                    throw new Error("A non-existing view was tried to insert " + name);
                }
                activeView = name;
            }
        },
        destroy: function () {
            viewCache.forEach(function (view) { return view === null || view === void 0 ? void 0 : view.destroy(); });
            viewContainerRef.clear();
        },
    };
}
function assertTemplate(property, templateRef) {
    var isTemplateRefOrNull = !!(!templateRef || templateRef.createEmbeddedView);
    if (!isTemplateRefOrNull) {
        throw new Error(property + " must be a TemplateRef, but received something else.");
    }
    return isTemplateRefOrNull;
}

/**
 * Noop Strategy
 *
 * This strategy is does nothing. It serves for debugging only
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `noop`      | ❌       | ❌             | ❌                | ❌         | ❌      |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy} - The calculated strategy
 *
 */
function createNoopStrategy() {
    return {
        name: 'noop',
        detectChanges: function () { },
        rxScheduleCD: function (o) { return o; },
        scheduleCD: function () { return new AbortController(); },
    };
}

/**
 * Native Strategy
 * @description
 *
 * - mFC - `cdRef.markForCheck`
 *
 * This strategy mirrors Angular's built-in `async` pipe.
 * This means for every emitted value `ChangeDetectorRef#markForCheck` is called.
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `native`    | ❌       | mFC           | ❌                | ❌         | ❌      |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy} - The calculated strategy
 *
 */
function createNativeStrategy(config) {
    var component = config.cdRef.context;
    return {
        name: 'native',
        detectChanges: function () { return config.cdRef.markForCheck(); },
        rxScheduleCD: function (o) { return o.pipe(tap(function () { return ɵmarkDirty(component); })); },
        scheduleCD: function () {
            ɵmarkDirty(component);
            return new AbortController();
        },
    };
}

var SchedulingPriority;
(function (SchedulingPriority) {
    SchedulingPriority["animationFrame"] = "animationFrame";
    SchedulingPriority["Promise"] = "Promise";
    SchedulingPriority["idleCallback"] = "idleCallback";
    SchedulingPriority["userBlocking"] = "userBlocking";
    SchedulingPriority["userVisible"] = "userVisible";
    SchedulingPriority["background"] = "background";
    SchedulingPriority["setInterval"] = "setInterval";
})(SchedulingPriority || (SchedulingPriority = {}));

var promiseDurationSelector = promiseTick();
/**
 * Experimental Local Strategies
 *
 * - ɵDC - `ɵdetectChanges`
 * - C - `Component`
 * - det - `cdRef.detach`
 * - ret - `cdRef.reattach`
 * - Pr - `Promise`
 * - aF - `requestAnimationFrame`
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `local`     | ✔        | ɵDC           | C + Pr           | aF         | ❌      |
 * | `detach`    | ✔ ️     | ret,ɵDC, det  | C + Pr           | aF         | ❌      |
 *
 */
function getLocalStrategies(config) {
    return {
        local: createLocalStrategy(config),
        detach: createDetachStrategy(config),
    };
}
/**
 *  Local Strategy
 *
 * This strategy is rendering the actual component and
 * all it's children that are on a path
 * that is marked as dirty or has components with `ChangeDetectionStrategy.Default`.
 *
 * As detectChanges has no coalescing of render calls
 * like `ChangeDetectorRef#markForCheck` or `ɵmarkDirty` has, so we have to apply our own coalescing, 'scoped' on
 * component level.
 *
 * Coalescing, in this very manner,
 * means **collecting all events** in the same
 * [EventLoop](https://developer.mozilla.org/de/docs/Web/JavaScript/EventLoop) tick, that would cause a re-render and
 * execute **re-rendering only once**.
 *
 * 'Scoped' coalescing, in addition, means **grouping the collected events by** a specific context.
 * E. g. the **component** from which the re-rendering was initiated.
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `local`     | ✔        | ɵDC           | C + Pr           | aF         | ❌      |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy} - The calculated strategy
 *
 */
function createLocalStrategy(config) {
    var component = config.cdRef.context;
    var priority = SchedulingPriority.animationFrame;
    var tick = priorityTickMap[priority];
    var renderMethod = function () {
        config.cdRef.detectChanges();
    };
    var behavior = function (o) {
        return o.pipe(coalesceWith(promiseDurationSelector, component), switchMap(function (v) { return tick.pipe(map(function () { return v; })); }), tap(renderMethod));
    };
    var scheduleCD = function () {
        return coalesceAndSchedule(renderMethod, priority, component);
    };
    return {
        name: 'local',
        detectChanges: renderMethod,
        rxScheduleCD: behavior,
        scheduleCD: scheduleCD,
    };
}
/**
 *  Detach Strategy
 *
 * This strategy is rendering the actual component and
 * all it's children that are on a path
 * that is marked as dirty or has components with `ChangeDetectionStrategy.Default`.
 *
 * As detectChanges has no coalescing of render calls
 * like `ChangeDetectorRef#markForCheck` or `ɵmarkDirty` has, so we have to apply our own coalescing, 'scoped' on
 * component level.
 *
 * Coalescing, in this very manner,
 * means **collecting all events** in the same
 * [EventLoop](https://developer.mozilla.org/de/docs/Web/JavaScript/EventLoop) tick, that would cause a re-render and
 * execute **re-rendering only once**.
 *
 * 'Scoped' coalescing, in addition, means **grouping the collected events by** a specific context.
 * E. g. the **component** from which the re-rendering was initiated.
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `detach`    | ✔ ️     | ret,ɵDC, det  | C + Pr           | aF         | ❌      |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy} - The calculated strategy
 *
 */
function createDetachStrategy(config) {
    var component = config.cdRef.context;
    var priority = SchedulingPriority.animationFrame;
    var tick = priorityTickMap[priority];
    var renderMethod = function () {
        config.cdRef.reattach();
        ɵdetectChanges(component);
        config.cdRef.detach();
    };
    var behavior = function (o) {
        return o.pipe(coalesceWith(promiseDurationSelector, component), switchMap(function (v) { return tick.pipe(map(function () { return v; })); }), tap(renderMethod));
    };
    var scheduleCD = function () {
        return coalesceAndSchedule(renderMethod, priority, component);
    };
    return {
        name: 'detach',
        detectChanges: renderMethod,
        rxScheduleCD: behavior,
        scheduleCD: scheduleCD,
    };
}

function getGlobalStrategies(config) {
    return {
        global: createGlobalStrategy(config),
    };
}
/**
 * Global Strategies
 *
 * - ɵMD - `ɵmarkDirty`
 * - C - `Component`
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `global`     | ✔        | ɵMD           | C + Pr          | ❌         | ❌      |
 *
 */
/**
 *
 * Global Strategy
 *
 * This strategy is rendering the application root and
 * all it's children that are on a path
 * that is marked as dirty or has components with `ChangeDetectionStrategy.Default`.
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `global`     | ✔        | ɵMD           | C + Pr          | ❌         | ❌      |
 *
 * @param config { RenderStrategyFactoryConfig } - The values this strategy needs to get calculated.
 * @return {RenderStrategy<T>} - The calculated strategy
 *
 */
function createGlobalStrategy(config) {
    var renderMethod = function () { return ɵmarkDirty(config.cdRef.context); };
    return {
        name: 'global',
        detectChanges: renderMethod,
        rxScheduleCD: function (o) { return o; },
        scheduleCD: function () {
            renderMethod();
            return new AbortController();
        },
    };
}

var DEFAULT_STRATEGY_NAME = 'local';
function getStrategies(config) {
    return __assign(__assign({ noop: createNoopStrategy(), native: createNativeStrategy(config) }, getGlobalStrategies(config)), getLocalStrategies(config));
}
/**
 * Strategies
 *
 * - mFC - `cdRef.markForCheck`
 * - dC - `cdRef.detectChanges`
 * - ɵMD - `ɵmarkDirty`
 * - ɵDC - `ɵdetectChanges`
 * - C - `Component`
 * - det - `cdRef.detach`
 * - ret - `cdRef.reattach`
 * - Pr - `Promise`
 * - aF - `requestAnimationFrame`
 *
 * | Name        | ZoneLess | Render Method | ScopedCoalescing | Scheduling | Chunked |
 * |-------------| ---------| --------------| ---------------- | ---------- |-------- |
 * | `noop`      | ❌       | ❌             | ❌               | ❌         | ❌       |
 * | `native`    | ❌       | mFC           | ❌                | ❌         | ❌      |
 * | `global`    | ✔        | ɵMD           | C + Pr           | ❌         | ❌      |
 * | `local`     | ✔        | ɵDC           | C + Pr           | aF         | ❌      |
 * | `detach`    | ✔ ️     | ret,ɵDC, det  | C + Pr           | aF         | ❌      |
 */

/**
 * @Directive LetDirective
 *
 * @description
 *
 * The `*rxLet` directive serves a convenient way of binding observables to a view context. Furthermore, it helps
 * you structure view-related models into view context scope (DOM element's scope).
 *
 * Under the hood, it leverages a `RenderStrategy` which in turn takes care of optimizing the change detection
 * of your component. The `LetDirective` will render its template and manage change detection after it got an initial value.
 * So if the incoming `Observable` emits its value lazily (e.g. data coming from `Http`), your template will be
 * rendered lazily as well. This can very positively impact the initial render performance of your application.
 *
 *
 * ### Problems with `async` and `*ngIf`
 *
 * In Angular, a way of binding an observable to the view could look like that:
 * ```html
 * <ng-container *ngIf="observableNumber$ | async as n">
 *   <app-number [number]="n"></app-number>
 *   <app-number-special [number]="n"></app-number-special>
 * </ng-container>
 * ```
 *
 * The problem is that `*ngIf` interferes with rendering and in case of a `0` (a falsy value) the component
 * would be hidden. This issue doesn't concern the `LetDirective`.
 *
 * The `AsyncPipe` relies on the Zone to be present - it doesn't really trigger change detection by itself.
 * It marks the component and its children as dirty waiting for the Zone to trigger change detection. So, in case
 * you want to create a zone-less application, the `AsyncPipe` won't work as desired. `LetDirective` comes
 * with its own strategies to manage change detection every time a new notification is sent from
 * the bound Observable.
 *
 *
 * ### Features of `*rxLet`
 *
 * Included features for `*rxLet`:
 * - binding is always present. (see "Problems with `async` and `*ngIf`" section below)
 * - it takes away the multiple usages of the `async` or `push` pipe
 * - a unified/structured way of handling null and undefined
 * - triggers change-detection differently if `zone.js` is present or not (`ChangeDetectorRef.detectChanges` or
 *   `ChangeDetectorRef.markForCheck`)
 * - triggers change-detection differently if ViewEngine or Ivy is present (`ChangeDetectorRef.detectChanges` or
 *   `ɵdetectChanges`)
 * - distinct same values in a row (`distinctUntilChanged` operator),
 * - display custom templates for different observable notifications (suspense, next, error, complete)
 *
 *
 * ### Binding an Observable and using the view context
 *
 * The `*rxLet` directive takes over several things and makes it more convenient and save to work with streams in the
 * template:
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 *
 * <ng-container *rxLet="observableNumber$ as n">
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * ```
 *
 * In addition to that it provides us information from the whole observable context.
 * We can track the observables:
 * - next value
 * - error occurrence
 * - complete occurrence
 *
 * ```html
 * <ng-container *rxLet="observableNumber$; let n; let e = $error, let c = $complete">
 *   <app-number [number]="n" *ngIf="!e && !c"></app-number>
 *   <ng-container *ngIf="e">
 *     There is an error: {{ e }}
 *   </ng-container>
 *   <ng-container *ngIf="c">
 *     Observable completed: {{ c }}
 *   </ng-container>
 * </ng-container>
 * ```
 *
 *
 * ### Using the template-binding
 *
 * You can also use template anchors and display template's content for different observable states:
 * - on complete
 * - on error
 * - on suspense - before the first value is emitted
 *
 * ```html
 * <ng-container
 *   *rxLet="
 *     observableNumber$;
 *     let n;
 *     error: error;
 *     complete: complete;
 *     suspense: suspense;
 *   "
 * >
 *   <app-number [number]="n"></app-number>
 * </ng-container>
 * <ng-template #error>ERROR</ng-template>
 * <ng-template #complete>COMPLETE</ng-template>
 * <ng-template #suspense>SUSPENSE</ng-template>
 * ```
 *
 * Internally, `*rxLet` is using a simple "view memoization" - it caches all anchored template references and re-uses
 * them whenever the observable notification (next/error/complete) is sent. Then, it only updates the context
 * (e.g. a value from the observable) in the view.
 *
 * @docsCategory LetDirective
 * @docsPage LetDirective
 * @publicApi
 */
var LetDirective = /** @class */ (function () {
    /**
     * @internal
     */
    function LetDirective(cdRef, nextTemplateRef, viewContainerRef) {
        var _this = this;
        this.nextTemplateRef = nextTemplateRef;
        this.viewContainerRef = viewContainerRef;
        this.subscription = new Subscription();
        this.resetObserver = {
            next: function () {
                _this.displayInitialView();
                _this.templateManager.updateViewContext({
                    $implicit: undefined,
                    rxLet: undefined,
                    $error: false,
                    $complete: false,
                });
            },
        };
        this.updateObserver = {
            next: function (value) {
                _this.templateManager.displayView('rxNext');
                _this.templateManager.updateViewContext({
                    $implicit: value,
                    rxLet: value,
                });
            },
            error: function (error) {
                // fallback to rxNext when there's no template for rxError
                _this.templateManager.hasTemplateRef('rxError')
                    ? _this.templateManager.displayView('rxError')
                    : _this.templateManager.displayView('rxNext');
                _this.templateManager.updateViewContext({
                    $error: true,
                });
            },
            complete: function () {
                // fallback to rxNext when there's no template for rxComplete
                _this.templateManager.hasTemplateRef('rxComplete')
                    ? _this.templateManager.displayView('rxComplete')
                    : _this.templateManager.displayView('rxNext');
                _this.templateManager.updateViewContext({
                    $complete: true,
                });
            },
        };
        this.displayInitialView = function () {
            // display "suspense" template if provided
            if (_this.templateManager.hasTemplateRef('rxSuspense')) {
                _this.templateManager.displayView('rxSuspense');
            }
        };
        this.strategies = getStrategies({ cdRef: cdRef });
        this.templateManager = createTemplateManager(this.viewContainerRef, {
            $implicit: undefined,
            rxLet: undefined,
            $error: false,
            $complete: false,
        });
        this.renderAware = createRenderAware({
            strategies: this.strategies,
            resetObserver: this.resetObserver,
            updateObserver: this.updateObserver,
        });
        this.renderAware.nextStrategy(DEFAULT_STRATEGY_NAME);
    }
    Object.defineProperty(LetDirective.prototype, "rxLet", {
        /**
         * @description
         * The Observable to be bound to the context of a template.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         *
         * @param potentialObservable
         */
        set: function (potentialObservable) {
            this.renderAware.nextPotentialObservable(potentialObservable);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "strategy", {
        /**
         * @description
         * The rendering strategy to be used when rendering with the reactive context within a template.
         * Use it to dynamically manage your rendering strategy. You can switch the strategies
         * imperatively (with a string) or by bounding an Observable.
         * The default strategy is `'local'`.
         *
         * @example
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *     <ng-container *rxLet="hero$; let hero; strategy: 'local'">
         *       <app-hero [hero]="hero"></app-hero>
         *     </ng-container>
         *   `
         * })
         * export class AppComponent {
         *   strategy = 'local';
         * }
         *
         * // OR
         *
         * \@Component({
         *   selector: 'app-root',
         *   template: `
         *     <ng-container *rxLet="hero$; let hero; strategy: strategy$">
         *       <app-hero [hero]="hero"></app-hero>
         *     </ng-container>
         *   `
         * })
         * export class AppComponent {
         *   strategy$ = new BehaviorSubject('local');
         * }
         *
         * @param strategy
         * @see {@link strategies}
         */
        set: function (strategy) {
            this.renderAware.nextStrategy(strategy || DEFAULT_STRATEGY_NAME);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "rxLetComplete", {
        /**
         * @description
         * A template to show if the bound Observable is in "complete" state.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero; complete: completeTemplate">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         * <ng-template #completeTemplate>
         *   <mat-icon>thumb_up</mat-icon>
         * </ng-template>
         *
         * @param templateRef
         */
        set: function (templateRef) {
            this.templateManager.addTemplateRef('rxComplete', templateRef);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "rxLetError", {
        /**
         * @description
         * A template to show if the bound Observable is in "error" state.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero; error: errorTemplate">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         * <ng-template #errorTemplate>
         *   <mat-icon>thumb_down</mat-icon>
         * </ng-template>
         *
         * @param templateRef
         */
        set: function (templateRef) {
            this.templateManager.addTemplateRef('rxError', templateRef);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(LetDirective.prototype, "rxLetSuspense", {
        /**
         * @description
         * A template to show before the first value is emitted from the bound Observable.
         *
         * @example
         * <ng-container *rxLet="hero$; let hero; suspense: suspenseTemplate">
         *   <app-hero [hero]="hero"></app-hero>
         * </ng-container>
         * <ng-template #suspenseTemplate>
         *   <mat-progress-spinner></mat-progress-spinner>
         * </ng-template>
         *
         * @param templateRef
         */
        set: function (templateRef) {
            this.templateManager.addTemplateRef('rxSuspense', templateRef);
        },
        enumerable: false,
        configurable: true
    });
    /**
     * @internal
     */
    LetDirective.ngTemplateContextGuard = function (dir, ctx) {
        return true;
    };
    /**
     * @internal
     */
    LetDirective.prototype.ngOnInit = function () {
        this.templateManager.addTemplateRef('rxNext', this.nextTemplateRef);
        this.displayInitialView();
        this.subscription = this.renderAware.subscribe();
    };
    /**
     * @internal
     */
    LetDirective.prototype.ngOnDestroy = function () {
        this.subscription.unsubscribe();
        this.templateManager.destroy();
    };
    LetDirective.ctorParameters = function () { return [
        { type: ChangeDetectorRef },
        { type: TemplateRef },
        { type: ViewContainerRef }
    ]; };
    LetDirective.decorators = [
        { type: Directive, args: [{
                    selector: '[rxLet]',
                    exportAs: 'renderNotifier',
                },] }
    ];
    LetDirective.ctorParameters = function () { return [
        { type: ChangeDetectorRef },
        { type: TemplateRef },
        { type: ViewContainerRef }
    ]; };
    LetDirective.propDecorators = {
        rxLet: [{ type: Input }],
        strategy: [{ type: Input, args: ['rxLetStrategy',] }],
        rxLetComplete: [{ type: Input }],
        rxLetError: [{ type: Input }],
        rxLetSuspense: [{ type: Input }]
    };
    return LetDirective;
}());

var EXPORTED_DECLARATIONS = [LetDirective];
var LetModule = /** @class */ (function () {
    function LetModule() {
    }
    LetModule.decorators = [
        { type: NgModule, args: [{
                    declarations: EXPORTED_DECLARATIONS,
                    exports: [EXPORTED_DECLARATIONS],
                },] }
    ];
    return LetModule;
}());

/**
 *
 * @description
 *
 * This function takes an elem and event and re-applies the listeners from the passed event to the
 * passed element with the zone un-patched version of it.
 *
 * @param elem {HTMLElement} - The elem to re-apply the listeners to.
 * @param event {string} - The name of the event from which to re-apply the listeners.
 *
 * @returns void
 */
function unpatchEventListener$1(elem, event) {
    var eventListeners = elem.eventListeners(event);
    // Return if no event listeners are present
    if (!eventListeners) {
        return;
    }
    var addEventListener = getZoneUnPatchedApi('addEventListener', elem).bind(elem);
    eventListeners.forEach(function (listener) {
        // Remove and reapply listeners with patched API
        elem.removeEventListener(event, listener);
        // Reapply listeners with un-patched API
        addEventListener(event, listener);
    });
}
function intersectionObserver(options) {
    var subject = new Subject();
    var observer = observerSupported()
        ? new IntersectionObserver(function (entries) {
            entries.forEach(function (entry) { return subject.next(entry); });
        }, options)
        : null;
    var entries$ = new Observable(function (subscriber) {
        subject.subscribe(subscriber);
        return function () {
            if (observer) {
                observer.disconnect();
            }
        };
    });
    return {
        entries$: entries$,
        observe: observer.observe,
        unobserve: observer.unobserve,
    };
}
var observerSupported = function () {
    return typeof window !== 'undefined'
        ? !!window.IntersectionObserver
        : false;
};
var ɵ0$1 = observerSupported;
var ViewportPrioDirective = /** @class */ (function () {
    function ViewportPrioDirective(el, letDirective) {
        var _this = this;
        this.el = el;
        this.letDirective = letDirective;
        this.entriesSubject = new Subject();
        this.entries$ = this.entriesSubject.pipe(mergeAll());
        this._viewportPrio = 'noop';
        /* @Input('viewport-prio')
         set viewportPrio(prio) {
           if (prio) {
             this._viewportPrio = prio || 'noop';
           }
         }*/
        this.observer = observerSupported()
            ? new IntersectionObserver(function (entries) { return _this.entriesSubject.next(entries); }, {
                threshold: 0,
            })
            : null;
        this.visibilityEvents$ = this.entries$.pipe(map(function (entry) {
            if (entry.intersectionRatio > 0) {
                return 'visible';
            }
            else {
                return 'invisible';
            }
        }));
    }
    ViewportPrioDirective.prototype.ngOnInit = function () {
        var _this = this;
        var letStrategyName$ = this.letDirective.renderAware.activeStrategy$.pipe(map(function (s) { return s.name; }), filter(function (name) { return name !== _this._viewportPrio; }));
        this.observer.observe(this.el.nativeElement);
        this.visibilityEvents$
            .pipe(withLatestFrom(letStrategyName$), map(function (_a) {
            var _b = __read(_a, 2), visibility = _b[0], strategyName = _b[1];
            return visibility === 'visible' ? strategyName : _this._viewportPrio;
        }), tap(function (strategyName) {
            _this.letDirective.strategy = strategyName;
            // render actual state on viewport enter
            // @TODO this doesnt catch unsubscribe (cant be cancelled)
            _this.letDirective.strategies[strategyName].scheduleCD();
        }))
            .subscribe();
    };
    ViewportPrioDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: LetDirective, decorators: [{ type: Optional }] }
    ]; };
    ViewportPrioDirective.decorators = [
        { type: Directive, args: [{
                    // tslint:disable-next-line:directive-selector
                    selector: '[viewport-prio]',
                },] }
    ];
    ViewportPrioDirective.ctorParameters = function () { return [
        { type: ElementRef },
        { type: LetDirective, decorators: [{ type: Optional }] }
    ]; };
    return ViewportPrioDirective;
}());

var DECLARATIONS$1 = [ViewportPrioDirective];
var ViewportPrioModule = /** @class */ (function () {
    function ViewportPrioModule() {
    }
    ViewportPrioModule.decorators = [
        { type: NgModule, args: [{
                    declarations: DECLARATIONS$1,
                    exports: DECLARATIONS$1,
                },] }
    ];
    return ViewportPrioModule;
}());

/**
 * envRunsIvy
 *
 * @description
 * Determines the used view engine of an Angular project is Ivy or not.
 * The check is done based on following table:
 * | render       | ViewEngine | ViewEngine | Ivy         | Ivy         |
 * | ------------ | ---------- | ---------- | ----------- | ----------- |
 * | **mode**     | prod       | dev        | prod        | dev         |
 * | **ng**       | present    | present    | `undefined` | present     |
 * | **ng.probe** | present    | present    | `undefined` | `undefined` |
 *
 *  So for Ivy we need to make sure that ng is undefined or,
 *  in case of dev environment, ng.probe is undefined.
 *
 * @return {boolean} - true if the used view engine is Ivy.
 *
 */
function isViewEngineIvy() {
    var ng = getGlobalThis().ng;
    // Is the global ng object is unavailable?
    // ng === undefined in Ivy production mode
    // View Engine has the ng object both in development mode and production mode.
    return (ng === undefined ||
        // in case we are in dev mode in ivy
        // `probe` property is available on ng object we use View Engine.
        ng.probe === undefined);
}

/**
 * @Pipe PushPipe
 *
 * @description
 *
 * The `push` pipe serves as a drop-in replacement for the `async` pipe.
 * It contains intelligent handling of change detection to enable us
 * running in zone-full as well as zone-less mode without any changes to the code.
 *
 * The current way of binding an observable to the view looks like that:
 *  ```html
 *  {{observable$ | async}}
 * <ng-container *ngIf="observable$ | async as o">{{o}}</ng-container>
 * <component [value]="observable$ | async"></component>
 * ```
 *
 * The problem is `async` pipe just marks the component and all its ancestors as dirty.
 * It needs zone.js microtask queue to exhaust until `ApplicationRef.tick` is called to render all dirty marked
 *     components.
 *
 * Heavy dynamic and interactive UIs suffer from zones change detection a lot and can
 * lean to bad performance or even unusable applications, but the `async` pipe does not work in zone-less mode.
 *
 * `push` pipe solves that problem.
 *
 * Included Features:
 *  - Take observables or promises, retrieve their values and render the value to the template
 *  - Handling null and undefined values in a clean unified/structured way
 *  - Triggers change-detection differently if `zone.js` is present or not (`detectChanges` or `markForCheck`)
 *  - Distinct same values in a row to increase performance
 *  - Coalescing of change detection calls to boost performance
 *
 * @usageNotes
 *
 * `push` pipe solves that problem. It can be used like shown here:
 * ```html
 * {{observable$ | push}}
 * <ng-container *ngIf="observable$ | push as o">{{o}}</ng-container>
 * <component [value]="observable$ | push"></component>
 * ```
 *
 * @publicApi
 */
var PushPipe = /** @class */ (function () {
    function PushPipe(cdRef) {
        var _this = this;
        this.resetObserver = {
            next: function () {
                _this.renderedValue = undefined;
            },
        };
        this.updateObserver = {
            next: function (value) { return (_this.renderedValue = value); },
        };
        this.RenderAware = createRenderAware({
            strategies: getStrategies({
                cdRef: cdRef,
            }),
            updateObserver: this.updateObserver,
            resetObserver: this.resetObserver,
        });
        this.subscription = this.RenderAware.subscribe();
    }
    PushPipe.prototype.transform = function (potentialObservable, config) {
        var strategy = config || DEFAULT_STRATEGY_NAME;
        this.RenderAware.nextStrategy(strategy);
        this.RenderAware.nextPotentialObservable(potentialObservable);
        return this.renderedValue;
    };
    PushPipe.prototype.ngOnDestroy = function () {
        this.subscription.unsubscribe();
    };
    PushPipe.ctorParameters = function () { return [
        { type: ChangeDetectorRef }
    ]; };
    PushPipe.decorators = [
        { type: Pipe, args: [{ name: 'push', pure: false },] }
    ];
    PushPipe.ctorParameters = function () { return [
        { type: ChangeDetectorRef }
    ]; };
    return PushPipe;
}());

var DECLARATIONS$2 = [PushPipe];
var PushModule = /** @class */ (function () {
    function PushModule() {
    }
    PushModule.decorators = [
        { type: NgModule, args: [{
                    declarations: DECLARATIONS$2,
                    imports: [],
                    exports: DECLARATIONS$2,
                },] }
    ];
    return PushModule;
}());

var TemplateModule = /** @class */ (function () {
    function TemplateModule() {
    }
    TemplateModule.decorators = [
        { type: NgModule, args: [{
                    exports: [LetModule, PushModule, UnpatchEventsModule, ViewportPrioModule],
                },] }
    ];
    return TemplateModule;
}());

/**
 * Generated bundle index. Do not edit.
 */

export { LetDirective, LetModule, PushModule, PushPipe, SchedulingPriority, TemplateModule, UnpatchEventsDirective, UnpatchEventsModule, ViewportPrioDirective, ViewportPrioModule, coalesceWith, getStrategies, getZoneUnPatchedApi, isNgZone, isViewEngineIvy, priorityTickMap, staticCoalesce };
//# sourceMappingURL=rx-angular-template.js.map
